from linkstore.linkstore import Linkstore
from linkstore.link_storage import SqliteConnectionFactory
from linkstore.factory import create_link, create_sqlite_link_storage_with_connection


def an_in_memory_sqlite_link_storage():
    return create_sqlite_link_storage_with_connection(SqliteConnectionFactory.create_in_memory())

def an_in_memory_sqlite_linkstore():
    return Linkstore(an_in_memory_sqlite_link_storage(), create_link)



import os
from os import path
import sqlite3

from .link import LinkRecord


class SqliteLinkStorage(object):
    def __init__(self, table_gateways):
        self._links_table = table_gateways['links']
        self._tags_table = table_gateways['tags']

    def get_all(self):
        return [
            LinkRecord(
                link_id,
                url,
                self._tags_table.get_tags_by_id(link_id),
                date_saved
            )
            for link_id, url, date_saved in self._links_table.get_all()
        ]

    def save(self, link):
        self._links_table.save(link.url, link.date)

        id_of_newly_created_link = self._links_table.get_id_by_url(link.url)
        self._tags_table.save_tags_for_link_with_id(id_of_newly_created_link, link.tags)

    def find_by_tag(self, tag):
        records_of_matching_links = []
        for link_id in self._tags_table.get_ids_of_links_with_tag(tag):
            url, date = self._links_table.get_url_and_date_by_id(link_id)

            records_of_matching_links.append(LinkRecord(link_id, url, self._tags_table.get_tags_by_id(link_id), date))

        return records_of_matching_links

    def replace_tag_in_link_with_url(self, url, tag_modification):
        id_of_relevant_link = self._links_table.get_id_by_url(url)
        self.replace_tag_in_link_with_id(id_of_relevant_link, tag_modification)

    def replace_tag_in_link_with_id(self, link_id, tag_modification):
        self._tags_table.replace_tag_in_link_with_id(link_id, tag_modification)

    def add_tag_to_link_with_url(self, url, tag):
        id_of_relevant_link = self._links_table.get_id_by_url(url)
        self.add_tag_to_link_with_id(id_of_relevant_link, tag)

    def add_tag_to_link_with_id(self, link_id, tag):
        self._tags_table.save_tags_for_link_with_id(link_id, (tag,))


class SqliteConnectionFactory(object):
    @staticmethod
    def create_autoclosing_on_disk():
        return AutoclosingSqliteConnection()

    @classmethod
    def create_in_memory(cls):
        connection_to_in_memory_database = sqlite3.connect(':memory:')
        cls._enable_enforcement_of_foreign_key_constraints(connection_to_in_memory_database)

        return connection_to_in_memory_database

    @staticmethod
    def _enable_enforcement_of_foreign_key_constraints(sqlite_connection):
        sqlite_connection.execute('pragma foreign_keys = on')

    @classmethod
    def create_on_disk(cls, data_directory):
        connection_to_on_disk_database = sqlite3.connect(data_directory.path_to_database_file)
        cls._enable_enforcement_of_foreign_key_constraints(connection_to_on_disk_database)

        return connection_to_on_disk_database


class SqliteTable(object):
    def __init__(self, connection_to_database):
        self._connection = connection_to_database

        self._set_up()

    def _set_up(self):
        with self._connection as connection:
            connection.execute(self.SQL_COMMAND_FOR_TABLE_CREATION)


class LinksTable(SqliteTable):
    SQL_COMMAND_FOR_TABLE_CREATION = '''
        create table if not exists links(
            link_id integer primary key
                not null,
            url
                unique
                not null,
            date_saved
                not null
        )
    '''

    def get_all(self):
        with self._connection as connection:
            return connection.execute('select link_id, url, date_saved from links').fetchall()

    def get_id_by_url(self, url):
        with self._connection as connection:
            row_of_link_with_given_url = connection.execute(
                'select link_id from links where url = ?',
                (url,)
            ).fetchone()
            desired_id = row_of_link_with_given_url[0]

            return desired_id

    def save(self, url, date):
        with self._connection as connection:
            connection.execute(
                'insert into links(url, date_saved) values(?, ?)',
                (url, date)
            )

    def get_url_and_date_by_id(self, link_id):
        with self._connection as connection:
            return connection.execute(
                'select url, date_saved from links where link_id = ?',
                (link_id,)
            ).fetchone()


class TagsTable(SqliteTable):
    SQL_COMMAND_FOR_TABLE_CREATION = '''
        create table if not exists tags(
            link_id
                not null,
            name
                not null,

            foreign key(link_id) references links(link_id)
                on delete restrict
                on update restrict
            )
    '''

    def save_tags_for_link_with_id(self, link_id, tags):
        with self._connection as connection:
            connection.executemany(
                'insert into tags(link_id, name) values(?, ?)',
                [(link_id, tag) for tag in tags]
            )

    def get_ids_of_links_with_tag(self, tag):
        with self._connection as connection:
            list_of_rows = connection.execute(
                'select link_id from tags where name = ?',
                (tag,)
            ).fetchall()

            return (link_id for (link_id,) in list_of_rows)

    def get_tags_by_id(self, link_id):
        with self._connection as connection:
            list_of_rows = connection.execute(
                'select name from tags where link_id = ?',
                (link_id,)
            ).fetchall()

            return tuple(tag for (tag,) in list_of_rows)

    def replace_tag_in_link_with_id(self, link_id, tag_modification):
        current_tag = tag_modification.keys()[0]
        new_tag = tag_modification.values()[0]

        with self._connection as connection:
            connection.execute(
                'update tags set name = ? where link_id = ? and name = ?',
                (new_tag, link_id, current_tag)
            )


class AutoclosingSqliteConnection(object):
    def __init__(self, provider_of_sqlite_connection=None):
        self._provider_of_sqlite_connection = provider_of_sqlite_connection if provider_of_sqlite_connection is not None \
            else ProviderOfConnectionToOnDiskSqliteDatabase()

    def __enter__(self):
        self._current_connection = self._provider_of_sqlite_connection.get()
        self._current_connection.__enter__()

        return self._current_connection

    def __exit__(self, type_, value, traceback):
        self._current_connection.__exit__(type_, value, traceback)
        self._current_connection.close()

        return False


class ProviderOfConnectionToOnDiskSqliteDatabase(object):
    def __init__(self):
        self._directory = ApplicationDataDirectory()

    def get(self):
        return SqliteConnectionFactory.create_on_disk(self._directory)


class ApplicationDataDirectory(object):
    @property
    def path(self):
        return path.expanduser('~/.linkstore/')

    @property
    def name_of_database_file(self):
        return 'linkstore.sqlite'

    @property
    def path_to_database_file(self):
        self._ensure_data_directory_exists()

        return path.join(self.path, self.name_of_database_file)

    def _ensure_data_directory_exists(self):
        if path.exists(self.path):
            return

        os.mkdir(self.path)



import subprocess
import shutil

from expects import expect, equal, be_empty, match, have_length, contain

from linkstore.link_storage import ApplicationDataDirectory


def invoke_cli(arguments):
    path_to_cli_binary = subprocess.check_output([ 'which', 'linkstore' ]).strip()

    try:
        output = subprocess.check_output(
            [ 'coverage', 'run', '--append' ] +
            [ path_to_cli_binary ] +
            arguments
        )
    except subprocess.CalledProcessError as error:
        return ExecutionResult(error.output, error.returncode)

    return ExecutionResult(output)

class ExecutionResult(object):
    def __init__(self, output, exit_code=0):
        self.lines_in_output = self._get_lines_from(output)
        self.exit_code = exit_code

    def _get_lines_from(self, output):
        if output == '':
            return []
        return output.strip().split('\n')


with description('the command-line application'):
    with context('when saving links'):
        with before.each:
            self.an_url = 'https://www.example.com/'

        with context('without a tag'):
            with it('fails'):
                execution_result = invoke_cli([ 'save', self.an_url ])

                expect(execution_result.exit_code).not_to(equal(0))

        with context('with one tag'):
            with it('does not output anything'):
                a_tag = 'favourites'

                execution_result = invoke_cli([ 'save', self.an_url, a_tag ])

                expect(execution_result.exit_code).to(equal(0))
                expect(execution_result.lines_in_output).to(be_empty)

        with context('with more than one tag'):
            with it('does not output anything'):
                a_tag = 'favourites'
                another_tag = 'another_tag'

                execution_result = invoke_cli([ 'save', self.an_url, a_tag, another_tag ])

                expect(execution_result.exit_code).to(equal(0))
                expect(execution_result.lines_in_output).to(be_empty)

    with context('when retrieving saved links'):
        with before.each:
            self.DATE_PATTERN = r'[0-9]{2}/[0-9]{2}/[0-9]{4}'

        with context('by tag'):
            with before.each:
                self.tag_filter = 'some_tag'
                links_to_save = [
                    ('some url',        self.tag_filter),
                    ('another url',     self.tag_filter),
                    ('yet another url', 'a_different_tag')
                ]

                for link in links_to_save:
                    invoke_cli([ 'save', link[0], link[1] ])
                self.saved_links = links_to_save


                self.execution_result = invoke_cli([ 'list', self.tag_filter ])

            with it('does not fail'):
                expect(self.execution_result.exit_code).to(equal(0))

            with it('outputs a line per matching link'):
                number_of_lines_in_output = len(self.execution_result.lines_in_output)
                number_of_matching_links = len([ link for link in self.saved_links if link[1] == self.tag_filter ])

                expect(number_of_lines_in_output).to(equal(number_of_matching_links))

            with context('output lines'):
                with it('have an id'):
                    NUMBER_AT_BEGINNING_OF_LINE_PATTERN = r'^\d+'

                    for line in self.execution_result.lines_in_output:
                        expect(line).to(match(NUMBER_AT_BEGINNING_OF_LINE_PATTERN))

                with it('have an URL'):
                    for line_number, line in enumerate(self.execution_result.lines_in_output):
                        url = self.saved_links[line_number][0]
                        expect(line).to(match(url))

                with it('have a date'):
                    for line in self.execution_result.lines_in_output:
                        expect(line).to(match(self.DATE_PATTERN))

                with it('have no tags'):
                    for line_number, line in enumerate(self.execution_result.lines_in_output):
                        tag = self.saved_links[line_number][1]

                        expect(line).not_to(match(tag))
                        expect(line).not_to(match(self.tag_filter))

        with context('all saved links'):
            with before.each:
                links_to_save = [
                    ('some url',        ('a tag',)),
                    ('another url',     ('first tag', 'second tag')),
                    ('yet another url', ('a_different_tag',))
                ]

                for link in links_to_save:
                    invoke_cli([ 'save', link[0] ] + list(link[1]))
                self.saved_links = links_to_save

                self.execution_result = invoke_cli([ 'list' ])

            with it('does not fail'):
                expect(self.execution_result.exit_code).to(equal(0))

            with it('outputs a line per saved link'):
                number_of_lines_in_output = len(self.execution_result.lines_in_output)
                number_of_saved_links = len(self.saved_links)

                expect(number_of_lines_in_output).to(equal(number_of_saved_links))

            with it('outputs the URL and tags of each saved link per line'):
                for line_number, line in enumerate(self.execution_result.lines_in_output):
                    url, tags = self.saved_links[line_number]

                    expect(line).to(match(url))
                    expect(line).to(match(self.DATE_PATTERN))

                    for tag in tags:
                        expect(line).to(match(tag))

    with context('when re-tagging links'):
        with before.each:
            self.an_url = 'https://www.example.com/'
            self.old_tag = 'favourites'
            invoke_cli([ 'save', self.an_url, self.old_tag ])

            self.new_tag = 'not-so-favourites'
            self.execution_result = invoke_cli([ 'retag', '1' ] + [ self.old_tag, self.new_tag ])

        with it('does not fail'):
            expect(self.execution_result.exit_code).to(equal(0))

        with it('does not output anything'):
            expect(self.execution_result.lines_in_output).to(be_empty)

        with it('removes the old tag'):
            links_matching_old_tag = invoke_cli([ 'list', self.old_tag ]).lines_in_output

            expect(links_matching_old_tag).to(be_empty)

        with it('adds the new tag'):
            links_matching_new_tag = invoke_cli([ 'list', self.new_tag ]).lines_in_output

            expect(links_matching_new_tag).to(have_length(1))
            expect(links_matching_new_tag[0]).to(match(self.an_url))

    with context('when tagging links'):
        with before.each:
            self.an_url = 'https://www.example.com/'
            self.old_tag = 'favourites'
            invoke_cli([ 'save', self.an_url, self.old_tag ])

            self.new_tag = 'very-favourites'
            self.execution_result = invoke_cli([ 'tag', '1', self.new_tag ])

        with it('does not fail'):
            expect(self.execution_result.exit_code).to(equal(0))

        with it('does not output anything'):
            expect(self.execution_result.lines_in_output).to(be_empty)

        with it('preserves the existing tag'):
            links_matching_old_tag = invoke_cli([ 'list', self.old_tag ]).lines_in_output

            expect(links_matching_old_tag).to(contain(match(self.an_url)))

        with it('adds the requested tag'):
            links_matching_new_tag = invoke_cli([ 'list', self.new_tag ]).lines_in_output

            expect(links_matching_new_tag).to(contain(match(self.an_url)))

    with after.each:
        shutil.rmtree(ApplicationDataDirectory().path)



from expects import expect, contain

from ..helpers import an_in_memory_sqlite_linkstore
from ..fixtures import one_link


with description('saving a link'):
    with it('is successfully saved'):
        a_link = one_link()
        linkstore = an_in_memory_sqlite_linkstore()

        linkstore.save(a_link)

        expect(linkstore.get_all()).to(contain(a_link))



from datetime import datetime


class Clock(object):
    def date_of_today(self):
        return self._convert_date_to_string_with_pattern(
            self._get_current_utc_date(),
            '%d/%m/%Y'
        )

    def _get_current_utc_date(self):
        return datetime.utcnow()

    def _convert_date_to_string_with_pattern(self, date, pattern):
        return date.strftime(pattern)



class Linkstore(object):
    def __init__(self, link_storage, link_creator):
        self._storage = link_storage
        self._link_creator = link_creator

    def save(self, link):
        self._storage.save(link)

    def find_by_tag(self, tag):
        return self._create_links_from_link_records(self._storage.find_by_tag(tag))

    def _create_links_from_link_records(self, link_records):
        return [
            self._link_creator(record)
            for record in link_records
        ]

    def get_all(self):
        return self._create_links_from_link_records(self._storage.get_all())

    def modify_tag(self, link, tag_modification):
        self._storage.replace_tag_in_link_with_url(link.url, tag_modification)

    def modify_tag_by_id(self, link_id, tag_modification):
        self._storage.replace_tag_in_link_with_id(link_id, tag_modification)

    def add_tag(self, link, tag):
        self._storage.add_tag_to_link_with_url(link.url, tag)

    def add_tag_by_id(self, link_id, tag):
        self._storage.add_tag_to_link_with_id(link_id, tag)



import sqlite3

from linkstore.link import Link
from linkstore.link_storage import SqliteLinkStorage, AutoclosingSqliteConnection, LinksTable, TagsTable

from expects import expect
from doublex import Stub, Spy, Mock
Dummy = Stub
from doublex_expects import have_been_called, have_been_called_with, have_been_satisfied_in_any_order, anything

from ..fixtures import a_tag_modification


with description('the SQLite link storage'):
    with context('when saving links'):
        with it('''reads the link's attributes'''):
            with Mock(Link) as link_mock:
                link_mock.url.returns('an url')
                link_mock.url.returns('an url')
                link_mock.tags.returns(('a tag', 'another tag'))
                link_mock.date.returns('a date')
            link_storage = SqliteLinkStorage({'links': Dummy(LinksTable), 'tags': Dummy(TagsTable)})

            link_storage.save(link_mock)

            expect(link_mock).to(have_been_satisfied_in_any_order)

    with context('when retrieving links by tag'):
        with it('asks the TagsTable for the ids of the links with the specified tag'):
            a_tag = 'favourites'
            with Spy(TagsTable) as tags_table_spy:
                tags_table_spy.get_ids_of_links_with_tag(a_tag).returns([])
            link_storage = SqliteLinkStorage({'links': Dummy(LinksTable), 'tags': tags_table_spy})

            link_storage.find_by_tag(a_tag)

            expect(tags_table_spy.get_ids_of_links_with_tag).to(
                have_been_called_with(a_tag).once
            )

    with context('when retrieving all links'):
        with it('asks the LinksTable for all its records'):
            with Spy(LinksTable) as links_table_spy:
                links_table_spy.get_all().returns([])
            link_storage = SqliteLinkStorage({'links': links_table_spy, 'tags': Dummy(TagsTable)})

            link_storage.get_all()

            expect(links_table_spy.get_all).to(
                have_been_called_with().once
            )

    with context('when replacing a tag of a link'):
        with before.each:
            self.tags_table_spy = Spy(TagsTable)
            self.link_storage = SqliteLinkStorage({'links': Dummy(LinksTable), 'tags': self.tags_table_spy})

        with context('when identifying the link with an url'):
            with it('asks the TagsTable to replace the tags'):
                an_url = 'an url'

                self.link_storage.replace_tag_in_link_with_url(an_url, a_tag_modification())

                expect(self.tags_table_spy.replace_tag_in_link_with_id).to(
                    have_been_called_with(anything, a_tag_modification()).once
                )

        with context('when identifying the link with an id'):
            with it('asks the TagsTable to replace the tags'):
                a_link_id = 32

                self.link_storage.replace_tag_in_link_with_id(a_link_id, a_tag_modification())

                expect(self.tags_table_spy.replace_tag_in_link_with_id).to(
                    have_been_called_with(a_link_id).once
                )

    with context('when adding a tag to a link'):
        with context('when identifying the link with an url'):
            with before.each:
                self.tags_table_spy = Spy(TagsTable)
                self.link_storage = SqliteLinkStorage({'links': Dummy(LinksTable), 'tags': self.tags_table_spy})

            with it('asks the TagsTable to add the tag'):
                an_url = 'an url'
                a_new_tag = 'a new tag'

                self.link_storage.add_tag_to_link_with_url(an_url, a_new_tag)

                expect(self.tags_table_spy.save_tags_for_link_with_id).to(
                    have_been_called_with(anything, (a_new_tag,)).once
                )


        with context('when identifying the link with an id'):
            with it('asks the TagsTable to add the tag'):
                a_link_id = 32
                a_new_tag = 'a new tag'

                self.link_storage.add_tag_to_link_with_id(a_link_id, a_new_tag)

                expect(self.tags_table_spy.save_tags_for_link_with_id).to(
                    have_been_called_with(a_link_id, (a_new_tag,)).once
                )


with description('the autoclosing SQLite connection'):
    with context('when used as a context manager'):
        with it('closes the connection after the `with` block is executed'):
            sqlite_connection = Spy(sqlite3.Connection)
            with Stub() as sqlite_connection_provider_stub:
                sqlite_connection_provider_stub.get().returns(sqlite_connection)

            with AutoclosingSqliteConnection(sqlite_connection_provider_stub):
                expect(sqlite_connection.close).not_to(have_been_called)

            expect(sqlite_connection.close).to(have_been_called_with().once)



import random

from linkstore.link import Link


def some_links():
    return map(create_link_from_tuple, data_for_some_links())

def create_link_from_tuple(link_tuple):
    return Link(*link_tuple)

def data_for_some_links():
    return [
        ('https://www.example.com/',        ('favourites',),                      '12/03/2008'   ),
        ('https://www.another-example.net', ('misc',),                            '35/56/89'     ),
        ('https://one-more.org',            ('bla', 'ble'),                       '789/23/677785'),
        ('an url',                          ('a tag',),                           '18/12/2015'   ),
        ('another url',                     ('another tag',),                     'a date'       ),
        ('one more url',                    ('a different tag',),                 'another date' ),
        ('and one more',                    ('with one tag', 'and one more tag'), 'whatever date')
    ]

def one_link():
    return random.choice(some_links())

def some_links_with_tags(tags):
    return [
        create_link_from_tuple((url, tags, date))
        for url, _, date in data_for_some_links()
    ]

def a_tag_modification():
    return {
        'old tag': 'new tag'
    }



from .linkstore import Linkstore
from .link_storage import SqliteLinkStorage, SqliteConnectionFactory, LinksTable, TagsTable


def create_linkstore_with_storage(storage):
    return Linkstore(
        storage,
        create_link
    )

def create_sqlite_link_storage():
    return create_sqlite_link_storage_with_connection(
        SqliteConnectionFactory.create_autoclosing_on_disk()
    )

def create_sqlite_link_storage_with_connection(connection):
    return SqliteLinkStorage({
        'links': LinksTable(connection),
        'tags': TagsTable(connection)
    })

def create_link(link_record):
    return link_record.toLink()



from expects import expect, equal

from linkstore.link import Link

from ..helpers import an_in_memory_sqlite_linkstore
from ..fixtures import some_links_with_tags, some_links


with description('retrieving links'):
    with before.each:
        self.linkstore = an_in_memory_sqlite_linkstore()

    with context('by one tag'):
        with it('returns the links originally saved with the target tag'):
            a_tag = 'the target tag'
            some_links_with_target_tag = some_links_with_tags((a_tag,))

            for link in some_links_with_target_tag:
                self.linkstore.save(link)

            all_links_with_target_tag = self.linkstore.find_by_tag(a_tag)
            expect(all_links_with_target_tag).to(equal(some_links_with_target_tag))

        with it('returns links originally saved with the target tag and some other tags'):
            a_tag = 'the target tag'
            some_links_with_target_tag_and_others = some_links_with_tags((a_tag, 'some other tag', 'yet some other tag'))

            for link in some_links_with_target_tag_and_others:
                self.linkstore.save(link)

            all_links_with_target_tag = self.linkstore.find_by_tag(a_tag)
            expect(all_links_with_target_tag).to(equal(some_links_with_target_tag_and_others))

        with it('''doesn't return links which weren't saved with the target tag'''):
            a_tag = 'the target tag'
            some_links_with_target_tag = some_links_with_tags((a_tag,))
            for link in some_links_with_target_tag:
                self.linkstore.save(link)
            a_different_link = Link('a different url', 'not the target tag', 'whatever date')

            self.linkstore.save(a_different_link)

            for link in self.linkstore.find_by_tag(a_tag):
                expect(link).not_to(equal(a_different_link))

    with context('all links'):
        with it('returns all links'):
            links_to_save = some_links()

            for link in links_to_save:
                self.linkstore.save(link)
            saved_links = links_to_save

            expect(self.linkstore.get_all()).to(equal(saved_links))



from expects import expect, contain

from ..fixtures import one_link
from ..helpers import an_in_memory_sqlite_linkstore


with description('changing tags'):
    with before.each:
        self.a_link = one_link()
        self.linkstore = an_in_memory_sqlite_linkstore()
        self.one_of_the_orignal_tags = self.a_link.tags[0]
        self.a_new_tag = 'a new tag'

        self.linkstore.save(self.a_link)

    with context('modifying tags'):
        with before.each:
            self.linkstore.modify_tag(self.a_link, {self.one_of_the_orignal_tags: self.a_new_tag})
            self.modified_link = self.linkstore.get_all()[0]

        with it('adds a new tag'):
            expect(self.modified_link.tags).to(contain(self.a_new_tag))

        with it('removes the old tag'):
            expect(self.modified_link.tags).not_to(contain(self.one_of_the_orignal_tags))

    with context('adding tags'):
        with before.each:
            self.linkstore.add_tag(self.a_link, self.a_new_tag)
            self.the_link = self.linkstore.get_all()[0]

        with it('preserves the original tag'):
            expect(self.the_link.tags).to(contain(self.one_of_the_orignal_tags))

        with it('adds the new tag'):
            expect(self.the_link.tags).to(contain(self.a_new_tag))



class Link(object):
    def __init__(self, url, tags, date):
        self._url = url
        self._tags = tags
        self._date = date

    @property
    def url(self):
        return self._url

    @property
    def tags(self):
        return self._tags

    @property
    def date(self):
        return self._date

    def __eq__(self, other):
        return all([
            self.url == other.url,
            self.tags == other.tags,
            self.date == other.date
        ])

    def __ne__(self, other):
        return not self.__eq__(other)

    __hash__ = None


class LinkRecord(object):
    def __init__(self, link_id, *parameters_for_link_constructor):
        self._id = str(link_id)
        self._link = Link(*parameters_for_link_constructor)

    @property
    def id(self):
        return self._id

    @property
    def url(self):
        return self._link.url

    @property
    def tags(self):
        return self._link.tags

    @property
    def date(self):
        return self._link.date

    def toLink(self):
        return self._link



from expects import expect, be_none
from doublex import Spy, Stub
from doublex_expects import have_been_called_with

from linkstore.link import Link
from linkstore.linkstore import Linkstore
from linkstore.link_storage import SqliteLinkStorage

from ..fixtures import data_for_some_links, a_tag_modification


with description('the link store'):
    with before.each:
        self.link_storage_spy = Spy(SqliteLinkStorage)
        self.link_creator_spy = Spy().link_creator_spy
        self.linkstore = Linkstore(self.link_storage_spy, self.link_creator_spy)

    with context('when adding a link'):
        with before.each:
            self.a_link = Stub(Link)

        with it('delegates to the storage'):
            self.linkstore.save(self.a_link)

            expect(self.link_storage_spy.save).to(
                have_been_called_with(self.a_link).once
            )

        with it('returns nothing'):
            return_value = self.linkstore.save(self.a_link)

            expect(return_value).to(be_none)

    with context('when retrieving links'):
        with context('by tag'):
            with before.each:
                self.a_tag = 'favourites'
                with self.link_storage_spy:
                    self.link_storage_spy.find_by_tag(self.a_tag).returns(data_for_some_links())

            with it('delegates to the storage'):
                self.linkstore.find_by_tag(self.a_tag)

                expect(self.link_storage_spy.find_by_tag).to(
                    have_been_called_with(self.a_tag).once
                )

            with it('calls the creator once per record returned by the storage'):
                self.linkstore.find_by_tag(self.a_tag)

                for link_record in data_for_some_links():
                    expect(self.link_creator_spy).to(have_been_called_with(link_record).once)

        with context('all links'):
            with context('without ids'):
                with before.each:
                    with self.link_storage_spy:
                        self.link_storage_spy.get_all().returns(data_for_some_links())

                with it('delegates to the storage'):
                    self.linkstore.get_all()

                    expect(self.link_storage_spy.get_all).to(
                        have_been_called_with().once
                    )

                with it('calls the creator once per record returned by the storage'):
                    self.linkstore.get_all()

                    for link_record in data_for_some_links():
                        expect(self.link_creator_spy).to(have_been_called_with(link_record).once)

            with context('with ids'):
                with before.each:
                    with self.link_storage_spy:
                        self.link_storage_spy.get_all_with_ids().returns(data_for_some_links())

    with context('when modifying a tag of a link'):
        with context('when identifying the link with a Link object'):
            with it('delegates to the storage'):
                a_link = Stub(Link)

                self.linkstore.modify_tag(a_link, a_tag_modification())

                expect(self.link_storage_spy.replace_tag_in_link_with_url).to(
                    have_been_called_with(a_link.url, a_tag_modification()).once
                )

        with context('when identifying the link with its id'):
            with it('delegates to the storage'):
                a_link_id = 45

                self.linkstore.modify_tag_by_id(a_link_id, a_tag_modification())

                expect(self.link_storage_spy.replace_tag_in_link_with_id).to(
                    have_been_called_with(a_link_id).once
                )

    with context('when adding a tag to a link'):
        with context('when identifying the link with a Link object'):
            with it('delegates to the storage'):
                a_link = Stub(Link)
                a_new_tag = 'a new tag'

                self.linkstore.add_tag(a_link, a_new_tag)

                expect(self.link_storage_spy.add_tag_to_link_with_url).to(
                    have_been_called_with(a_link.url, a_new_tag).once
                )

        with context('when identifying the link with its id'):
            with it('delegates to the storage'):
                a_link_id = 45
                a_new_tag = 'a new tag'

                self.linkstore.add_tag_by_id(a_link_id, a_new_tag)

                expect(self.link_storage_spy.add_tag_to_link_with_id).to(
                    have_been_called_with(a_link_id, a_new_tag).once
                )



import click
from click import group, argument

from . import factory
from .link import Link
from .clock import Clock


link_storage = factory.create_sqlite_link_storage()
linkstore = factory.create_linkstore_with_storage(link_storage)
clock = Clock()


@group()
def linkstore_cli():
    pass

@linkstore_cli.command()
@argument('url')
@argument('tags', nargs=-1, required=True)
def save(url, tags):
    linkstore.save(Link(url, tags, clock.date_of_today()))


@linkstore_cli.command()
@argument('tag_filter', required=False)
def list(tag_filter):
    if tag_filter is None:
        print_all_links()
    else:
        print_without_tags_links_tagged_with(tag_filter)

def print_all_links():
    for link_record in link_storage.get_all():
        print('  |  '.join([
            link_record.id,
            link_record.url,
            link_record.date,
            '#' + ', #'.join(link_record.tags)
        ]))

def print_without_tags_links_tagged_with(tag_filter):
    for link_record in link_storage.find_by_tag(tag_filter):
        print('  |  '.join([
            link_record.id,
            link_record.url,
            link_record.date
        ]))


@linkstore_cli.command()
@argument('link_id', type=click.INT)
@argument('current_tag')
@argument('new_tag')
def retag(link_id, current_tag, new_tag):
    linkstore.modify_tag_by_id(link_id, {current_tag: new_tag})


@linkstore_cli.command()
@argument('link_id', type=click.INT)
@argument('new_tag')
def tag(link_id, new_tag):
    linkstore.add_tag_by_id(link_id, new_tag)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


setup(
    name = 'linkstore',
    version = '0.0.7',
    description = 'A dead-simple bookmarking CLI application',
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
        'License :: OSI Approved :: MIT License'
    ],
    keywords = 'link hyperlink store storage',
    author = u'Ángel Sanz',
    author_email = 'angelsanzgit@gmail.com',
    url = 'https://github.com/angelsanz/linkstore',
    license = 'MIT',
    packages = find_packages(),
    include_package_data = True,
    zip_safe = False,
    install_requires = [
        requirement.strip() for requirement in open('requirements.txt').readlines()
    ],
    entry_points = {
        'console_scripts': [
            'linkstore = linkstore.cli:linkstore_cli'
        ]
    }
)



from expects import expect, equal

from linkstore.link import Link


with description('Link'):
    with context('is a value object'):
        with it('respects equality'):
            expect(
                Link('an url', ('a tag',), 'a date')
            ).to(equal(
                Link('an url', ('a tag',), 'a date')
            ))

        with it('respects non-equality'):
            pairs_of_different_links = [
                (Link('an url', ('a tag',), 'a date'), Link('another url', ('a tag',),       'a date')),
                (Link('an url', ('a tag',), 'a date'), Link('an url',      ('another tag',), 'a date')),
                (Link('an url', ('a tag',), 'a date'), Link('an url',      ('a tag',),       'another date')),
                (Link('an url', ('a tag',), 'a date'), Link('another url', ('another tag',), 'a date')),
                (Link('an url', ('a tag',), 'a date'), Link('another url', ('a tag',),       'another date')),
                (Link('an url', ('a tag',), 'a date'), Link('an url',      ('another tag',), 'another date')),
                (Link('an url', ('a tag',), 'a date'), Link('another url', ('another tag',), 'another date'))
            ]

            for a_link, a_different_link in pairs_of_different_links:
                expect(a_link).not_to(equal(a_different_link))
